/*
	         Java RayTracer
	     By Dean Camera, 2008
	http://www.fourwalledcubicle.com
*/

public class PrimSphere implements ITraceableObject
{
	private static final Vector3D SPHERELATLONGCROSS = Vector3D.CrossProduct(Vector3D.UNITVECTORY, Vector3D.UNITVECTORX);

	private float     Radius;
	private float     RadiusSquared;
	private Vector3D  Location;
	private ITexture  Texture;
	private IMaterial Material;

	public PrimSphere(float Radius, Vector3D Location, ITexture Texture, IMaterial Material)
	{
		this.Radius   = Radius;
		this.Location = Location;
		this.Texture  = Texture;
		this.Material = Material;
		
		RadiusSquared = Radius * Radius;
	}

	public Vector3D GetLocation()
	{
		return Location;
	}
	
	public IMaterial GetMaterial()
	{
		return Material;
	}

	public void SetLocation(Vector3D NewLocation)
	{
		Location = NewLocation;
	}

	public float GetCollisionDistance(Ray CurrentRay)
	{
		Vector3D SphereCenterToRay = Vector3D.Subtract(Location, CurrentRay.GetOrigin());

		float B = Vector3D.DotProduct(SphereCenterToRay, CurrentRay.GetDirection());

		// If ray is pointing away from the sphere, no intersection
		if (B < 0)
			return Float.MAX_VALUE;

		float C = Vector3D.DotProduct(SphereCenterToRay, SphereCenterToRay) - RadiusSquared;

		// Don't allow the intersection to occur inside the sphere
		if (C < 0)
			return Float.MAX_VALUE;

		float D = (B * B) - C;

		// If discriminant is less than zero, no intersection
		if (D < 0)
			return Float.MAX_VALUE;

		// Calculate parametric value
		return (B - (float)Math.sqrt(D));
	}
	
	public CollisionDetails GetCollisionDetails(Ray CurrentRay, float CollisionDistance)
	{
		// Create a point at the place of intersection relative to the eye
		Vector3D IntersectionPoint = Vector3D.Multiply(CurrentRay.GetDirection(), CollisionDistance).Add(CurrentRay.GetOrigin());
		
		// Calculate normal to the point of intersection
		Vector3D Normal = Vector3D.Subtract(IntersectionPoint, Location).UnitVector();

		// Calculate UV mapping coordinates
		float Phi   = (float)Math.acos(-Vector3D.DotProduct(Vector3D.UNITVECTORY, Normal));
		float Theta = (float)((Math.acos(Vector3D.DotProduct(Normal, Vector3D.UNITVECTORX) / Math.sin(Phi))) / (2 * Math.PI));
		float V = (float)(Phi / Math.PI);
		float U = (Vector3D.DotProduct(SPHERELATLONGCROSS, Normal) > 0) ? Theta : (1 - Theta);

		// Pass back the collision details to the caller
		return new CollisionDetails(this, IntersectionPoint, CollisionDistance, Texture.GetUVMappedColor(U, V), Normal);
	}
}